//
// Copyright: Avnet 2020
// Created by Nik Markovic <nikola.markovic@avnet.com> on 6/24/21.
//
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "iotconnect_common.h"
#include "iotconnect.h"

#include "app_config.h"
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>

// windows compatibility
#if defined(_WIN32) || defined(_WIN64)
#define F_OK 0
#include <Windows.h>
#include <io.h>
int usleep(unsigned long usec) {
    Sleep(usec / 1000);
    return 0;
}
#define access    _access_s
#else
#include <unistd.h>
#endif

/***                Definitions                 ***/
#define APP_VERSION "00.01.00"

/***              Global Variables              ***/
int fd_P1_0 = 0, fd_P5_1 = 0;

/***              Private Functions             ***/

static void on_connection_status(IotConnectConnectionStatus status) {
    // Add your own status handling
    switch (status) {
    case IOTC_CS_MQTT_CONNECTED:
        printf("IoTConnect Client Connected\n");
        break;
    case IOTC_CS_MQTT_DISCONNECTED:
        printf("IoTConnect Client Disconnected\n");
        break;
    default:
        printf("IoTConnect Client ERROR\n");
        break;
    }
}

static void command_status(IotclEventData data, bool status, const char* command_name, const char* message) {
    const char* ack = iotcl_create_ack_string_and_destroy_event(data, status, message);
    printf("command: %s status=%s: %s\n", command_name, status ? "OK" : "Failed", message);
    printf("Sent CMD ack: %s\n", ack);
    iotconnect_sdk_send_ack_packet(ack);

    fd_P1_0 = open("/sys/class/gpio/P1_0/value", O_WRONLY);
    if (fd_P1_0 == -1) {
        perror("Unable to open /sys/class/gpio/P1_0/value");
        exit(1);
    }

    fd_P5_1 = open("/sys/class/gpio/P5_1/value", O_WRONLY);
    if (fd_P5_1 == -1) {
        perror("Unable to open /sys/class/gpio/P5_1/value");
        exit(1);
    }

    if ((strcmp(command_name, "Green_LED ON") == 0) || (strcmp(command_name, "Green_LED on") == 0)) {
        if (write(fd_P1_0, "1", 1) != 1) {
            perror("Error writing to /sys/class/gpio/P1_0/value");
            exit(1);
        }
        else {
            printf("Green LED ON !!!!\n");
        }
    }

    else if ((strcmp(command_name, "Green_LED OFF") == 0) || (strcmp(command_name, "Green_LED off") == 0)) {
        if (write(fd_P1_0, "0", 1) != 1) {
            perror("Error writing to /sys/class/gpio/P1_0/value");
            exit(1);
        }
        else {
            printf("Green LED OFF !!!!\n");
        }
    }

    else if ((strcmp(command_name, "Red_LED ON") == 0) || (strcmp(command_name, "Red_LED on") == 0)) {
        if (write(fd_P5_1, "1", 1) != 1) {
            perror("Error writing to /sys/class/gpio/P1_0/value");
            exit(1);
        }
        else {
            printf("Red LED ON !!!!\n");
        }
    }

    else if ((strcmp(command_name, "Red_LED OFF") == 0) || (strcmp(command_name, "Red_LED off") == 0)) {
        if (write(fd_P5_1, "0", 1) != 1) {
            perror("Error writing to /sys/class/gpio/P1_0/value");
            exit(1);
        }
        else {
            printf("Red LED OFF !!!!\n");
        }
    }
    else {
        perror("Command Not matched as Defined  !!!!\n");
    }
    close(fd_P1_0);
    close(fd_P5_1);
    free((void*)ack);
}

static void on_command(IotclEventData data) {
    char* command = iotcl_clone_command(data);
    if (NULL != command) {
        command_status(data, true, command, "Not implemented");
        free((void*)command);
    }
    else {
        command_status(data, true, "?", "Internal error");
    }
}

static bool is_app_version_same_as_ota(const char* version) {
    return strcmp(APP_VERSION, version) == 0;
}

static bool app_needs_ota_update(const char* version) {
    return strcmp(APP_VERSION, version) < 0;
}

static void on_ota(IotclEventData data) {
    const char* message = NULL;
    char* url = iotcl_clone_download_url(data, 0);
    bool success = false;
    if (NULL != url) {
        printf("Download URL is: %s\n", url);
        const char* version = iotcl_clone_sw_version(data);
        if (is_app_version_same_as_ota(version)) {
            printf("OTA request for same version %s. Sending success\n", version);
            success = true;
            message = "Version is matching";
        }
        else if (app_needs_ota_update(version)) {
            printf("OTA update is required for version %s.\n", version);
            success = false;
            message = "Not implemented";
        }
        else {
            printf("Device firmware version %s is newer than OTA version %s. Sending failure\n", APP_VERSION,
                version);
            // Not sure what to do here. The app version is better than OTA version.
            // Probably a development version, so return failure?
            // The user should decide here.
            success = false;
            message = "Device firmware version is newer";
        }

        free((void*)url);
        free((void*)version);
    }
    else {
        // compatibility with older events
        // This app does not support FOTA with older back ends, but the user can add the functionality
        const char* command = iotcl_clone_command(data);
        if (NULL != command) {
            // URL will be inside the command
            printf("Command is: %s\n", command);
            message = "Old back end URLS are not supported by the app";
            free((void*)command);
        }
    }
    const char* ack = iotcl_create_ack_string_and_destroy_event(data, success, message);
    if (NULL != ack) {
        printf("Sent OTA ack: %s\n", ack);
        iotconnect_sdk_send_ack_packet(ack);
        free((void*)ack);
    }
}


static void publish_telemetry() {
    IotclMessageHandle msg = iotcl_telemetry_create();

    // Optional. The first time you create a data point, the current timestamp will be automatically added
    // TelemetryAddWith* calls are only required if sending multiple data points in one packet.
    iotcl_telemetry_add_with_iso_time(msg, iotcl_iso_timestamp_now());
    iotcl_telemetry_set_string(msg, "version", APP_VERSION);
    iotcl_telemetry_set_number(msg, "cpu", 3.123); // test floating point numbers
    iotcl_telemetry_set_number(msg, "temp", 30);
    const char* str = iotcl_create_serialized_string(msg, false);
    iotcl_telemetry_destroy(msg);
    printf("Sending: %s\n", str);
    iotconnect_sdk_send_packet(str); // underlying code will report an error
    iotcl_destroy_serialized(str);
}

/*! Gpio unexport - Private Functions
 *  \The method will unexport the dedicated gpios from /sys/class/gpio path.
 *  \If unexport fails function will exit()
 */
static void gpio_unexport() {
    fd_P1_0 = open("/sys/class/gpio/unexport", O_WRONLY);
    if (fd_P1_0 == -1) {
        perror("Unable to open /sys/class/gpio/unexport");
        exit(1);
    }
    if (write(fd_P1_0, "368", 3) != 3) {
        perror("Error writing to /sys/class/gpio/unexport");
        exit(1);
    }
    close(fd_P1_0);

    fd_P5_1 = open("/sys/class/gpio/unexport", O_WRONLY);
    if (fd_P5_1 == -1) {
        perror("Unable to open /sys/class/gpio/unexport");
        exit(1);
    }
    if (write(fd_P5_1, "401", 3) != 3) {
        perror("Error writing to /sys/class/gpio/unexport");
        exit(1);
    }
    close(fd_P5_1);
}

/*! Gpio export direction out - Private Functions
 *  \The method will export the dedicated gpios and set the direction to out.
 *  \If export or out direction fails function will exit()
 */
static void gpio_export_direction_out(void) {
    fd_P1_0 = open("/sys/class/gpio/export", O_WRONLY);
    fd_P5_1 = open("/sys/class/gpio/export", O_WRONLY);
    if (fd_P1_0 == -1 || fd_P5_1 == -1) {
        perror("Unable to open /sys/class/gpio/export");
        exit(1);
    }
    if ((write(fd_P1_0, "368", 3) != 3) || (write(fd_P5_1, "401", 3) != 3)) {
        close(fd_P1_0);
        close(fd_P5_1);

        gpio_unexport();//If writing to the gpio file fails must unexport all gpio and start fresh export again.
        fd_P1_0 = open("/sys/class/gpio/export", O_WRONLY);
        if (fd_P1_0 == -1) {
            perror("Unable to open /sys/class/gpio/export");
            exit(1);
        }
        if (write(fd_P1_0, "368", 3) != 3) {
            perror("Error writing to P1_0 /sys/class/gpio/export");
            exit(1);
        }
        close(fd_P1_0);

        fd_P5_1 = open("/sys/class/gpio/export", O_WRONLY);
        if (fd_P5_1 == -1) {
            perror("Unable to open /sys/class/gpio/export");
            exit(1);
        }
        if (write(fd_P5_1, "401", 3) != 3) {
            perror("Error writing to P5_1 /sys/class/gpio/export");
            exit(1);
        }
        close(fd_P5_1);
    }

    // Set the pin to be an output by writing "out" to /sys/class/gpio/P1_0/direction
    fd_P1_0 = open("/sys/class/gpio/P1_0/direction", O_WRONLY);
    if (fd_P1_0 == -1) {
        perror("Unable to open /sys/class/gpio/P1_0/direction");
        exit(1);
    }
    if (write(fd_P1_0, "out", 3) != 3) {
        perror("Error writing to /sys/class/gpio/P1_0/direction");
        exit(1);
    }
    close(fd_P1_0);

    // Set the pin to be an output by writing "out" to /sys/class/gpio/P5_1/direction
    fd_P5_1 = open("/sys/class/gpio/P5_1/direction", O_WRONLY);
    if (fd_P5_1 == -1) {
        perror("Unable to open /sys/class/gpio/P5_1/direction");
        exit(1);
    }
    if (write(fd_P5_1, "out", 3) != 3) {
        perror("Error writing to /sys/class/gpio/P5_1/direction");
        exit(1);
    }
    close(fd_P5_1);
}

int main(int argc, char* argv[]) {
    if (access(IOTCONNECT_SERVER_CERT, F_OK) != 0) {
        fprintf(stderr, "Unable to access IOTCONNECT_SERVER_CERT. "
            "Please change directory so that %s can be accessed from the application or update IOTCONNECT_CERT_PATH\n",
            IOTCONNECT_SERVER_CERT);
    }

    if (IOTCONNECT_AUTH_TYPE == IOTC_AT_X509) {
        if (access(IOTCONNECT_IDENTITY_CERT, F_OK) != 0 ||
            access(IOTCONNECT_IDENTITY_KEY, F_OK) != 0
            ) {
            fprintf(stderr, "Unable to access device identity private key and certificate. "
                "Please change directory so that %s can be accessed from the application or update IOTCONNECT_CERT_PATH\n",
                IOTCONNECT_SERVER_CERT);
        }
    }

    IotConnectClientConfig* config = iotconnect_sdk_init_and_get_config();
    config->cpid = IOTCONNECT_CPID;
    config->env = IOTCONNECT_ENV;
    config->duid = IOTCONNECT_DUID;
    config->auth_info.type = IOTCONNECT_AUTH_TYPE;
    config->auth_info.trust_store = IOTCONNECT_SERVER_CERT;
    config->discovery_host = IOTCONNECT_DISCOVERY_HOSTNAME;

    if (config->auth_info.type == IOTC_AT_X509) {
        config->auth_info.data.cert_info.device_cert = IOTCONNECT_IDENTITY_CERT;
        config->auth_info.data.cert_info.device_key = IOTCONNECT_IDENTITY_KEY;
    }
    else if (config->auth_info.type == IOTC_AT_TPM) {
        config->auth_info.data.scope_id = IOTCONNECT_SCOPE_ID;
    }
    else if (config->auth_info.type == IOTC_AT_SYMMETRIC_KEY) {
        config->auth_info.data.symmetric_key = IOTCONNECT_SYMMETRIC_KEY;
    }
    else if (config->auth_info.type != IOTC_AT_TOKEN) { // token type does not need any secret or info
     // none of the above
        fprintf(stderr, "IOTCONNECT_AUTH_TYPE is invalid\n");
        return -1;
    }


    config->status_cb = on_connection_status;
    config->ota_cb = on_ota;
    config->cmd_cb = on_command;

    // Initialization export the gpio and sets the direction of gpio's as out.
    gpio_export_direction_out();

    // run a dozen connect/send/disconnect cycles with each cycle being about a minute
    for (int j = 0; j < 10; j++) {
        int ret = iotconnect_sdk_init();
        if (0 != ret) {
            fprintf(stderr, "IoTConnect exited with error code %d\n", ret);
            return ret;
        }

        // send 100 messages
        for (int i = 0; iotconnect_sdk_is_connected() && i < 100; i++) {
            publish_telemetry();
            // repeat approximately evey ~5 seconds
            for (int k = 0; k < 500; k++) {
                iotconnect_sdk_receive();
                usleep(10000); // 10ms
            }
        }
        iotconnect_sdk_disconnect();
    }
    //unexport the all the gpios
    gpio_unexport();
    return 0;
}
